import pandas as pd


def merge_groups_dummies(data_set_path):

    dummies_path = data_set_path.replace('.csv', '_dummies_groupscore.csv')
    dummies_df =pd.read_csv(dummies_path,index_col=[0,1])
    dummies_df.drop(['is_churn'],axis=1,inplace=True)

    groups_path = data_set_path.replace('.csv', '_nocat_groupscore.csv')
    groups_df = pd.read_csv(groups_path,index_col=[0,1])

    merged_df= groups_df.merge(dummies_df,left_index=True,right_index=True)
    save_path = data_set_path.replace('.csv', '_groupscore.csv')
    merged_df.to_csv(save_path)
    print('Saved merged group score + dummy dataset ' + save_path)

    standard_group_metrics = pd.read_csv(data_set_path.replace('.csv', '_nocat_groupmets.csv'),index_col=0)
    dummies_group_metrics = pd.read_csv(data_set_path.replace('.csv', '_dummies_groupmets.csv'),index_col=0)
    merged_col_df = standard_group_metrics.append(dummies_group_metrics)
    merged_col_df.to_csv(data_set_path.replace('.csv', '_groupmets.csv'))



import hydra
import random
import numpy as np
import os
import pandas as pd
import psycopg2 as post
import tempfile

from datetime import timedelta, datetime
from dateutil import parser
from dateutil.relativedelta import *
from joblib import Parallel, delayed
from filelock import FileLock
from math import ceil
from omegaconf import DictConfig, OmegaConf
from postgres import Postgres
from shutil import copyfile

from fightchurn.churnsim.behavior import LogNormalBehaviorModel
from fightchurn.churnsim.churndb import drop_schema, setup_churn_db
from fightchurn.churnsim.utility import UtilityModel
from fightchurn.churnsim.customer import  Customer

class ChurnSimulation:

    def __init__(self, args):
        '''
        Creates the behavior/utility model objects, sets internal variables to prepare for simulation, and creates
        the database connection

        For a detailed explanation the ChurnSim report:
        https://papers.ssrn.com/sol3/papers.cfm?abstract_id=4540160
        - Section 3.2.1 Behavior Model
        - Section 3.3 Simulation Algorithm
        - Section 3.4.2 Product Channels
        - Section 3.4.5 Customer Location

        :param model: name of the behavior/utility model parameters
        :param start: start date for simulation
        :param end: end date for simulation
        :param init_customers: how many customers to create at start date
        '''
        self.args = args
        self.model_name = args.model
        self.start_date = parser.parse(args.start_date).date()
        self.end_date = parser.parse(args.end_date).date()
        self.init_customers=args.init_customers
        self.monthly_growth_rate = args.growth_rate
        self.devmode= args.dev
        self.n_parallel = args.n_parallel
        self.utility_dist = []
        print(f'Simulating with {self.n_parallel} parallel processes...')
        self.util_mod=UtilityModel(self.model_name,args)

        self.min_age = args.min_age
        self.max_age = args.max_age
        self.age_satisfaction_scale = args.age_satisfy

        self.behavior_models = {}
        self.model_list = []
        self.population_percents = self.args.population

        for version in self.population_percents.keys():
            behave_mod=LogNormalBehaviorModel(self.model_name, exp_base=self.args.behave_exp_base,
                                              random_seed= args.random_seed, version= version)
            self.behavior_models[behave_mod.version]=behave_mod
            self.model_list.append(behave_mod)

        local_dir = os.path.join(f'{os.path.abspath(os.path.dirname(__file__))}','conf')

        self.save_path = os.path.join(os.getenv('CHURN_OUT_DIR') , self.model_name )

        plans_path =  os.path.join(local_dir, self.model_name + '_plans.csv')
        self.plans = pd.read_csv(plans_path,index_col=0)
        os.makedirs(self.save_path, exist_ok=True)
        copy_path = os.path.join(self.save_path,  f'{self.model_name}_plans.csv')
        copyfile(plans_path, copy_path)

        # self.plans = self.plans.sort_values('mrr',ascending=True) # Make sure its sorted by increasing MRR - wait, why?
        add_on_file = os.path.join(local_dir, self.model_name + '_addons.csv')
        if os.path.exists(add_on_file):
            self.add_ons = pd.read_csv(add_on_file)
            copy_path = os.path.join(self.save_path,  f'{self.model_name}_addons.csv')
            copyfile(add_on_file, copy_path)
        else:
            self.add_ons = pd.DataFrame()
        self.util_mod.setExpectations(self.behavior_models,self.population_percents)

        self.country_lookup=self.args.country

        self.tmp_sub_file_name = os.path.join(tempfile.gettempdir(),f'{self.model_name}_tmp_sub.csv')
        self.tmp_event_file_name=os.path.join(tempfile.gettempdir(),f'{self.model_name}_tmp_event.csv')

        copyfile(plans_path, copy_path)
        arg_path = os.path.join(self.save_path, f'{self.model_name}_conf.yaml')
        OmegaConf.save(self.args,arg_path)


    def con_string(self):
        return f"postgresql://{os.environ.get('CHURN_DB_HOST','localhost')}/{os.environ['CHURN_DB']}?user={os.environ['CHURN_DB_USER']}&password={os.environ['CHURN_DB_PASS']}"


    def sim_rate_debug_query(self):

        file_root = os.path.abspath(os.path.dirname(__file__))
        with open(os.path.join(file_root,'schema','churn_by_plan.sql'), 'r') as sqlfile:
            sql = sqlfile.read().replace('\n', ' ')
        sql = sql.replace('%schema',self.model_name)
        with FileLock(Customer.ID_LOCK_FILE):
            db = Postgres(self.con_string())
            res = pd.DataFrame(db.all(sql))
            print(res)


    def remove_tmp_files(self):
        '''
        Remove temp files. Runs at the start in case a previous run failed, and at the end.
        :return:
        '''
        try:
            os.remove(Customer.ID_LOCK_FILE)
        except OSError:
            pass
        try:
            os.remove(Customer.ID_FILE)
        except OSError:
            pass


    def simulate_customer(self, start_of_month):
        '''
        Simulate one customer collecting its events and subscriptions. This is  core inner loop of the simulation.
        For a detailed explanation see Section 3.3, Simulation Algorithm, of the ChurnSim report:
        https://papers.ssrn.com/sol3/papers.cfm?abstract_id=4540160
        - Also see section 3.4.5 Customer Location regarding the country

        This function has the core interaction between the simulation objects.  Customer is created from the behavior
        model, and picking a random start date within the month.  Then the customer objects simulates the events for
        the month, and the utility model determines if there is a churn based on the simulated event counts.

        :param start_of_month:
        :return: the new customer object it contains the events and subscriptions
        '''
        customer_model = self.behavior_models[np.random.choice(list(self.population_percents.keys()),
                                          p= list(self.population_percents.values()))]
        new_customer=customer_model.generate_customer(start_of_month,args=self.args)

        customer_country = np.random.choice(list(self.country_lookup.keys()),p=list(self.country_lookup.values()))
        new_customer.country = customer_country

        plans_to_use = self.plans
        available_periods = None
        if 'bill_period' in self.plans.columns.values:
            available_periods = sorted(self.plans['bill_period'].unique())
            # For the first month in a simulation, only allow 1 month bill periods
            #  - otherwise there are artificial cycles of lots of people re-newing on the 3,6,12 months
            if start_of_month==self.start_date:
                min_period=self.plans['bill_period'].min()
                plans_to_use =  self.plans[self.plans['bill_period']==min_period]

        new_customer.pick_initial_plan(plans_to_use, self.add_ons, available_periods)

        # Pick a random start date for the subscription within the month
        end_range = start_of_month + relativedelta(months=+1)
        customer_start = start_of_month + timedelta(days=random.randrange((end_range-start_of_month).days))
        next_renewal = customer_start + relativedelta(months=new_customer.bill_period)

        def add_customer_subscriptions(this_month, next_renewal):
            plan_units, plan_quantity = self.get_unit_quantity(new_customer.plan)
            new_customer.subscriptions.append( (new_customer.plan, this_month,
                                                next_renewal,
                                                new_customer.base_mrr, plan_quantity, plan_units, new_customer.bill_period, new_customer.discount ))
            for add_on in new_customer.add_ons.iterrows():
                add_units, add_quantity = self.get_unit_quantity(add_on[1]['plan'])
                new_customer.subscriptions.append( (add_on[1]['plan'],this_month,next_renewal, add_on[1]['mrr'],
                                                    add_quantity,add_units, new_customer.bill_period, new_customer.discount) )

        add_customer_subscriptions(customer_start, next_renewal)

        churned = False
        churn_intent_count = 0
        num_bill_periods = 1
        num_months = 0
        num_months_this_bill_period = 0
        while not churned:
            this_month = customer_start if num_months == 0 else next_month
            num_months += 1
            num_months_this_bill_period += 1
            next_month=customer_start+relativedelta(months=num_months)
            month_event_count = new_customer.generate_events(this_month,next_month)
            _ = self.util_mod.utility_function(month_event_count,new_customer)
            churn_intent = False
            if self.args.acausal_churn > 0:
                if random.uniform(0,1) < self.args.acausal_churn:
                    churn_intent = True
            if not churn_intent:
                churn_intent =self.util_mod.simulate_churn(month_event_count,new_customer)
            if churn_intent:
                churn_intent_count += 1
            # exit if the simulation is over
            if next_month > self.end_date:
                break
            # check for churn, upgrade/downgrade on renewal date, make new subscriptions if not churned
            elif next_month >= next_renewal:
                # churn at the end of the term if they wanted to churn just once (every month for monthly)
                if churn_intent > 0:
                    churned = True
                if not churned:
                    churn_intent_count = 0
                    num_months_this_bill_period = 0
                    bill_period_change = self.util_mod.simulate_upgrade_downgrade(month_event_count,new_customer,self.plans,self.add_ons)
                    if bill_period_change:
                        customer_start = next_month
                        num_months=0
                        num_bill_periods=1
                    else:
                        num_bill_periods += 1
                    next_renewal = customer_start + relativedelta(months=new_customer.bill_period * num_bill_periods)
                    add_customer_subscriptions(next_month,next_renewal)
            elif (churn_intent_count > 1 and 2 <= new_customer.bill_period <= 6) or \
                (churn_intent_count > 2 and 6 < new_customer.bill_period):
                # churn mid term on 2 churns for quarterly/bi-annual, 3 churns for annual
                churned=True
                old_last = new_customer.subscriptions[-1]
                new_last = (old_last[0],old_last[1],next_month,old_last[3],old_last[4],old_last[5], old_last[6], old_last[7])
                new_customer.subscriptions[-1]= new_last

        return new_customer

    def get_unit_quantity(self,plan):
        if plan in self.plans.index.values:
            if self.plans.shape[1]>2:
                limit_col = self.plans.columns.values[2]
                return  limit_col,  self.plans.loc[plan,limit_col]

        if len(self.add_ons)> 0 and plan in self.add_ons['plan'].values:
            if self.add_ons.shape[1]>3:
                add_on = self.add_ons[self.add_ons['plan']==plan]
                for limit_col in self.add_ons.columns.values[3:]:
                    if add_on[limit_col].values[0]>0:
                        return limit_col, add_on[limit_col].values[0]

        return None, None

    def create_customers_for_month(self,month_date,n_to_create):
        '''
        Creates all the customers for one month, by calling simulate_customer and copy_customer_to_database in a loop.
        :param month_date: the month start date
        :param n_to_create: number of customers to create within that month
        :return:
        '''

        def create_one_customer():
            customer = self.simulate_customer(month_date)
            self.copy_customer_to_database(customer)
            if self.devmode and customer.id> 0 and (customer.id % round(self.init_customers / 10)) == 0:
                self.sim_rate_debug_query()
            res = np.append(customer.utility_contribs, customer.current_utility)
            return res

        contribs = Parallel(n_jobs=self.n_parallel)(delayed(create_one_customer)() for i in range(n_to_create))

        if self.devmode:
            self.utility_dist.extend(contribs)
            util_names = list(self.util_mod.behave_names)
            util_names.extend(['mrr','utility'])
            utility_obs = pd.DataFrame(self.utility_dist, columns=util_names)
            utility_obs.to_csv( os.path.join(self.save_path,f'utility_contribs_{datetime.strftime(month_date,"%Y%m%d")}.csv'))
            utility_desc=pd.DataFrame(utility_obs.describe())
            utility_desc.to_csv( os.path.join(self.save_path,f'utility_contribs_{datetime.strftime(month_date,"%Y%m%d")}_summary.csv'))

    def copy_customer_to_database(self,customer):
        '''
        Copy one customers data to the database, by first writing it to temp files and then using the sql COPY command
        :param customer: a Customer object that has already had its simulation run
        :return:
        '''
        sub_file_name = self.tmp_sub_file_name.replace('.csv', f'{customer.id}.csv')
        event_file_name = self.tmp_event_file_name.replace('.csv', f'{customer.id}.csv')
        db = Postgres(self.con_string())

        with open(sub_file_name, 'w') as tmp_file:
            for s in customer.subscriptions:
                # plan name, start, end, mrr, quantity, units, billing period, discount
                tmp_file.write(f'{customer.id},{s[0]},{s[1]},{s[2]},{s[3]},'
                               f'{s[4] if s[4] is not None else "NULL"},{s[5] if s[5] is not None else "NULL"},'
                               f'{s[6]},{s[7]}\n')
        with open(event_file_name, 'w') as tmp_file:
            for e in customer.events:
                tmp_file.write(f'{customer.id},{e[0]},{e[1]},{e[2]},{e[3] if e[3] is not None else "NULL"}\n') # event time, event type id, user id, value

        sql = "INSERT INTO {}.account VALUES({},'{}','{}',{})".format(self.model_name, customer.id, customer.channel,
                                                                customer.date_of_birth.isoformat(),
                                                                'NULL' if customer.country == 'None' else "'{}'".format(
                                                                    customer.country))

        db.run(sql)

        con = post.connect( database= os.environ['CHURN_DB'],
                                 user= os.environ['CHURN_DB_USER'],
                                 password=os.environ['CHURN_DB_PASS'],
                                 host=os.environ.get('CHURN_DB_HOST','localhost'))
        cur = con.cursor()

        sql = "COPY %s.subscription FROM STDIN USING DELIMITERS ',' WITH NULL AS 'NULL'" % (self.model_name)
        with open(sub_file_name, 'r') as f:
            cur.copy_expert(sql, f)
        con.commit()

        sql = "COPY %s.event FROM STDIN USING DELIMITERS ',' WITH NULL AS 'NULL'" % (self.model_name)
        with open(event_file_name, 'r') as f:
            cur.copy_expert(sql, f)
        con.commit()
        con.close()

        os.remove(event_file_name)
        os.remove(sub_file_name)


    def truncate_old_sim(self,force):
        '''
        Removes an old simulation from the database, if it already exists for this model
        :return: True if is safe to proceed (no data or data removed); False means old data not removed
        '''
        db= Postgres(self.con_string())

        exists = db.one(f"SELECT exists(select schema_name FROM information_schema.schemata WHERE schema_name = '{self.model_name}')")
        if exists:
            if not force:
                print('TRUNCATING *Events/Metrics & Subscriptions/Observations* in schema -> %s <-  ...' % self.model_name)
                if input("are you sure? (enter %s to proceed) " % self.model_name) != self.model_name:
                    return False
            drop_schema(self.model_name)
            setup_churn_db(self.model_name)
            return True
        else:
            setup_churn_db(self.model_name)
            return True


    def run_simulation(self, force=False):
        '''
        Simulation main function. First it prepares the database by truncating any old events and subscriptions, and
        inserting the event types into the database.  Next it creeates the initial customers by calling
        create_customers_for_month, and then it advances month by month adding new customers (also using
        create_customers_for_month.)  The number of new customers for each month is determined from the growth rate.
        Note that churn is not handled at this level, but is modeled at the customer level.

        For a detailed explanation see Section 3.3, Simulation Algorithm, of the ChurnSim report:
        https://papers.ssrn.com/sol3/papers.cfm?abstract_id=4540160
        :return:
        '''

        # database setup
        if not self.truncate_old_sim(force):
            return
        self.remove_tmp_files()

        # Any model can insert the event types
        db = Postgres(self.con_string())
        self.behavior_models[next(iter(self.behavior_models))].insert_event_types(self.model_name,db)

        # Initial customer count
        print('\nCreating %d initial customers for month of %s @ %s' % (self.init_customers,self.start_date,datetime.now().strftime("%Y-%m-%d %H:%M:%S")))
        self.create_customers_for_month(self.start_date,self.init_customers)
        print('Created %d initial customers for start date %s @ %s' % (self.init_customers,str(self.start_date),datetime.now().strftime("%Y-%m-%d %H:%M:%S")))

        # Advance to additional months
        next_month=self.start_date+relativedelta(months=+1)
        n_to_add = int(ceil( self.init_customers* self.monthly_growth_rate))  # number of new customers in first month
        while next_month < self.end_date:
            print('\nCreating %d new customers for month of %s @ %s:' % (n_to_add,next_month,datetime.now().strftime("%Y-%m-%d %H:%M:%S")))
            self.create_customers_for_month(next_month,n_to_add)
            print('Created %d new customers for month %s @ %s\n' % (n_to_add,str(next_month),datetime.now().strftime("%Y-%m-%d %H:%M:%S")))
            next_month=next_month+relativedelta(months=+1)
            n_to_add = int(ceil( n_to_add * (1.0+self.monthly_growth_rate))) # increase the new customers by growth

        self.remove_tmp_files()
        self.sim_rate_debug_query()


@hydra.main(version_base=None, config_path="conf", config_name="socialnet7")
def run_churn_simulation(cfg : DictConfig):
    print(OmegaConf.to_yaml(cfg))
    if cfg.random_seed is not None:
        random.seed(cfg.random_seed) # for random
    churn_sim = ChurnSimulation(cfg)
    churn_sim.run_simulation(force=cfg.force)


if __name__ == "__main__":
    run_churn_simulation()



import pandas as pd
import os

from fightchurn.listings.chap10.listing_10_2_category_churn_cohorts import category_churn_summary, category_churn_plot, prepare_category_data



def grouped_category_cohorts(data_set_path, cat_col, groups):

    churn_data = prepare_category_data(data_set_path,cat_col)

    group_cat_col = group_category_column(churn_data,cat_col,groups)

    summary = category_churn_summary(churn_data,group_cat_col,data_set_path)

    category_churn_plot(group_cat_col, summary, data_set_path)

def group_category_column(df, cat_col, group_dict):
    group_lookup = {value: key for key in group_dict.keys() for value in group_dict[key]}
    group_cat_col = cat_col + '_group'
    df[group_cat_col] = df[cat_col].apply(lambda x: group_lookup[x] if x in group_lookup else x)
    df.drop(cat_col,axis=1,inplace=True)
    return group_cat_col


'''
Creates a schema and the tables for Fighting Churn With Data in a PostgreSQL database.

The schema name to create is specified by the constant `schema_name` - change that to
create a different schema.

The database and login information are in environment variables:
CHURN_DB
CHURN_DB_USER
CHURN_DB_PASS

The SQL statements to create the tables are the adjacent directory ../schema

'''

from postgres import Postgres
import os
import sys

def drop_schema(schema_name):
    con_string = f"postgresql://{os.environ.get('CHURN_DB_HOST','localhost')}/{os.environ['CHURN_DB']}?user={os.environ['CHURN_DB_USER']}&password={os.environ['CHURN_DB_PASS']}"
    db = Postgres(con_string)
    db.run(f'DROP SCHEMA IF EXISTS {schema_name} CASCADE;')

def setup_churn_db(schema_name):

    con_string = f"postgresql://{os.environ.get('CHURN_DB_HOST','localhost')}/{os.environ['CHURN_DB']}?user={os.environ['CHURN_DB_USER']}&password={os.environ['CHURN_DB_PASS']}"
    db = Postgres(con_string)

    tables=['event','subscription','event_type','metric','metric_name','active_period','observation','active_week','account']

    print('Creating schema %s...' % schema_name)
    db.run('CREATE SCHEMA IF NOT EXISTS %s;' % schema_name)

    for t in tables:
        file_root = os.path.abspath(os.path.dirname(__file__))
        with open(os.path.join(file_root,'schema',f'create_{t}.sql')) as sqlfile:
            sql = sqlfile.read().replace('\n', ' ')
        sql=sql.replace('x.','%s.' % schema_name)
        print('Creating table %s' % t)
        db.run(sql)

if __name__ == "__main__":
    schema_name='socialnet7'
    if len(sys.argv) >= 2:
        schema_name = sys.argv[1]
    drop_schema(schema_name)
    setup_churn_db(schema_name)



from fightchurn.listings.chap5.listing_5_2_dataset_stats import dataset_stats
from fightchurn.listings.chap8.listing_8_6_clipped_scores import clipped_scores
from fightchurn.listings.chap6.listing_6_4_find_metric_groups import find_metric_groups
from fightchurn.listings.chap6.listing_6_3_apply_metric_groups import apply_metric_groups

def prepare_data(data_set_path):
    dataset_stats(data_set_path)
    clipped_scores(data_set_path)
    find_metric_groups(data_set_path)
    apply_metric_groups(data_set_path)



import pandas as pd
import os
import pickle
import matplotlib.pyplot as plt
from fightchurn.listings.chap8.listing_8_4_rescore_metrics import reload_churn_data

def churn_forecast(data_set_path,model_name='logreg_model'):

    pickle_path = data_set_path.replace('.csv', '_{}.pkl'.format(model_name))
    assert os.path.isfile(pickle_path), 'You must run listing 8.2 to save a logistic regression model first'
    with open(pickle_path, 'rb') as fid:
        logreg_model = pickle.load(fid)

    current_score_df = reload_churn_data(data_set_path,'current_groupscore','8.4',is_customer_data=True)

    predictions = logreg_model.predict_proba(current_score_df.to_numpy())

    predict_df = pd.DataFrame(predictions, index=current_score_df.index, columns=['churn_prob', 'retain_prob'])
    forecast_save_path = data_set_path.replace('.csv', '_current_predictions.csv')
    print('Saving results to %s' % forecast_save_path)
    predict_df.to_csv(forecast_save_path, header=True)

    forecast_histogram(data_set_path,predict_df)

def forecast_histogram(data_set_path,predict_df,ext='reg'):
    plt.figure(figsize=[6,4])
    n, bins,_ = plt.hist(predict_df['churn_prob'].values,bins=20)
    plt.xlabel('Churn Probability')
    plt.ylabel('# of Accounts')
    plt.title('Histogram of Active Customer Churn Probability ({})'.format(ext))
    plt.grid()
    plt.savefig(data_set_path.replace('.csv', '_{}_churnhist.png'.format(ext)), format='png')
    plt.close()
    hist_df=pd.DataFrame({'n':n,'bins':bins[1:]})
    hist_df.to_csv(data_set_path.replace('.csv', '_current_churnhist.csv'))



import numpy as np
import os
import pandas as pd

from math import  exp
from omegaconf import OmegaConf
from random import uniform

from fightchurn.churnsim.customer import Customer

class UtilityModel:

    def __init__(self,name, args):
        '''
        This class calculates the churn probability for a customer based on their event counts.  Its called a "Utility
        Model" to mean utility in the economic sense: How much a good or service to satisfies one or more needs or
        wants of a consumer. So how likely the customer is to churn depends on how much utility they get, which is
        based on how many events they have from the behavior model.

        The parameters of the utility model are loaded from a file.  The current implementation loads a file with the
        same form as the behavior model: a vector and a matrix.  The number of these must match the number of
        behaviors in the behavior model. The vector is for calculating multiplicative utility: Each element is multiplied
        by the number of events to get the model utility from those behaviors. The matrix is unused as of this time,
        but the idea was that a second term could be added with the matrix defining utility interactions between the
        behaviors. The utility is calculated in the function `utility_function`

        The churn probability is a sigmoidal function based on utility - see the function `simulate_churn`.

        :param name:
        :param churn_rate: Target churn rate for calibration
        :param behavior_model: The behavior model that this utility function works withy
        '''
        self.name=name
        self.args = args
        self.contrib_scale =self.args.util_contrib_scale
        self.util_df = pd.DataFrame([list(args.utility.keys()),list(args.utility.values())]).transpose().set_index(0)
        self.util_df.columns=['util']
        self.util_df=self.util_df.astype({'util':float})
        if 'users' in self.util_df.index.values:
            assert self.util_df.index.get_loc('users')==self.util_df.shape[0]-1, "users should be last in utility list (if included)"

        self.utility_weights=self.util_df['util']
        self.mrr_utility_cost=self.utility_weights.loc['mrr']
        if self.mrr_utility_cost > 0:
            print(f"*** WARNING: MRR Should have a non-positive utility impact, but found {self.mrr_utility_cost}")
        self.utility_weights = self.utility_weights.drop(['mrr'],axis=0)
        self.behave_names=self.utility_weights.index.values
        self.transition_df = pd.json_normalize(OmegaConf.to_container(self.args.transition).values(),max_level=1)
        self.transition_df.index = list(self.args.transition.keys())

        # Setup in setExpectations,below
        self.behave_means = None
        self.expected_contributions = None
        self.avg_n_user = 0

        save_path = os.path.join(os.getenv('CHURN_OUT_DIR') , self.name )
        os.makedirs(save_path, exist_ok=True)
        copy_path = os.path.join(save_path,  f'{name}_utility.csv')
        self.util_df.to_csv(copy_path)
        copy_path = os.path.join(save_path,  f'{name}_updownchurn.csv')
        self.transition_df.to_csv(copy_path)

    def setExpectations(self, chan_mod_dict, model_weights):
        """
        Calculates the average utility contribution from each behavior, for use in the utility function.
        For a detailed explanation see  the ChurnSim report:
        https://papers.ssrn.com/sol3/papers.cfm?abstract_id=4540160
        - Section 3.2.2, Utility Model
        - Section 3.4.2, Product Channels

        :param chan_mod_dict: dictionary from channel name to the associated behavior model
        :param model_weights: probability for each channel to be chosen
        :return:
        """
        assert sum(model_weights.values())==1.0, "Model weights should sum to 1.0"
        n_behaviors = len(self.behave_names)
        self.behave_means = np.zeros(n_behaviors)
        for bemod in chan_mod_dict.values():
            underlying_behaviors = Customer.get_underlying_behaviors(bemod.behave_names)
            value_behaviors = {v: Customer.get_behavior_under_value(v,bemod.behave_names) for v in Customer.get_valued_behaviors(bemod.behave_names) }
            assert n_behaviors == len(underlying_behaviors)
            assert all(self.behave_names == underlying_behaviors)
            weight = model_weights[bemod.version]

            # behaviors with values multiply average number times average value
            model_means = bemod.behave_means[underlying_behaviors]
            for vb in value_behaviors.keys():
                model_means.loc[value_behaviors[vb]]=model_means.loc[value_behaviors[vb]]*bemod.behave_means[vb]

            # expected values of non-user behaviors are per-user, so multiply to get the expected total
            if 'users' in bemod.behave_means.index.values:
                model_means[model_means.index!='users'] = model_means[model_means.index!='users']*model_means[model_means.index=='users'].values[0]
                self.avg_n_user = self.avg_n_user + weight* bemod.behave_means['users']
            else:
                self.avg_n_user = self.avg_n_user + weight* 1

            self.behave_means = self.behave_means + weight * np.array(model_means)


        self.expected_contributions = self.behave_means * self.utility_weights.values


    def utility_function(self,event_counts,customer):
        '''
        Utility calculation for a customer:
        1. Take the ratios of the customer's event counts to the mean event counts
        2. Apply an exponential growth function modeling hedonic adaptation
        3. Add the term for MRR
        4. Apply customer satisfaction propensity


        For a detailed explanation see the ChurnSim report:
        https://papers.ssrn.com/sol3/papers.cfm?abstract_id=4540160
        - Section 3.2.2, Utility Model
        - Section 3.4.3, Customer Satisfiability Coefficient
        - Section 3.5.3, Product Plans, MRR
        
        :param event_counts:
        :return:
        '''
        contrib_ratios = event_counts / self.behave_means
        utility_contribs = self.expected_contributions * (1.0 - np.exp(-self.contrib_scale*contrib_ratios))
        mrr_utility = customer.mrr* self.mrr_utility_cost
        if customer.discount > 0:
            # utility due to getting a discount: -1 so that with mrr_utility_cost this is a positive contribution
            mrr_utility += -1* customer.mrr* self.mrr_utility_cost * customer.discount * self.args.discount_satisfy
        utility_contribs =  np.append(utility_contribs,mrr_utility)
        utility = np.sum(utility_contribs)

        if customer.satisfaction_propensity != 1.0:
            multiplier = customer.satisfaction_propensity if utility > 0.0  else (1.0/customer.satisfaction_propensity)
        else:
            multiplier = 1.0
        utility *= multiplier

        customer.current_utility = utility
        customer.utility_contribs = utility_contribs
        return utility

    def transition_probility(self,u,trans):
        '''
        Transition probability function for upsell, downsell and churn.
        For a detailed explanation see the ChurnSim report:
        https://papers.ssrn.com/sol3/papers.cfm?abstract_id=4540160
        - Section 3.2.3, Churn Model
        - Section 3.5.4, Upgrade & Downgrade

        :param u: The utility value
        :param trans: Which transition (string)
        :return: The probability
        '''
        offset = self.transition_df.loc[trans,'offset']
        scale = self.transition_df.loc[trans,'scale']
        # Clip exponent to prevent overflow errors in extreme cases
        prob=1.0/(1.0 + exp( np.clip(-1.*scale * u + offset, -40.0, 40.0)))
        if trans in ['downsell','churn']:
            prob = 1.0 - prob
        return prob


    def churn_probability(self,u):
        churn_prob = self.transition_probility(u,'churn')
        return churn_prob

    def downgrade_probability(self,u):
        down_prob= self.transition_probility(u,'downsell')
        return down_prob


    def uprade_probability(self,u):
        up_prob = self.transition_probility(u,'upsell')
        return up_prob

    def simulate_churn(self,event_counts,customer):
        '''
        Simulates one customer churn, given a set of event counts.  The retention probability is a sigmoidal function
        in the utility, and the churn probability is 100% minus retention. The return value is a binary indicating
        churn or no churn, by comparing a uniform random variable on [0,1] to the churn probability.
        :param event_counts:
        :return:
        '''
        return uniform(0, 1) < self.churn_probability(customer.current_utility)

    def simulate_upgrade_downgrade(self,event_counts,customer,plans, add_ons):
        '''
        Determine one customer's upgrade, downgrade, changes in add-ons and changes in billing period.
        For a detailed explanation see the ChurnSim report:
        https://papers.ssrn.com/sol3/papers.cfm?abstract_id=4540160
        - Section 3.5.4, Upgrade & Downgrade
        - Section 3.5.5, Add-on Products
        - Section 3.5.6, Billing Periods

        :param event_counts:
        :param customer:
        :param plans:
        :return:
        '''
        current_plan = np.where(plans.index.values==customer.plan)[0][0]
        upgrade_probability = self.uprade_probability(customer.current_utility)
        downgrade_probability = self.downgrade_probability(customer.current_utility)

        changed_plan=False
        changed_add_ons=False
        changed_bill_period=False
        if uniform(0, 1) < upgrade_probability:
            near_limit=None
            if plans.shape[1]>2:
                for limited in plans.columns.values[2:]:
                    current_limit=plans.iloc[current_plan][limited]
                    max_limit=plans[limited].max()
                    customer_rate = customer.get_behavior_rate(limited)
                    if customer_rate>= 0.5* current_limit and max_limit>current_limit:
                        near_limit = limited
                        break

            if near_limit is not None:
                # upgrade to a higher limit plan with the same billing period
                same_period_plans = plans[plans['bill_period']==customer.bill_period]
                # sorting so the customer takes the next higher limit plan
                higher_limit_plans =  same_period_plans[same_period_plans[limited]>current_limit].sort_values(by=limited)
                if len(higher_limit_plans)>0:
                    new_plan_name = higher_limit_plans.index.values[0]
                    customer.set_plan(plans,plan_name=new_plan_name)
                    changed_plan=True
            # if there was no suitable upgrade, try to increase the billing period
            if not changed_plan and 'bill_period' in plans.columns.values:
                # upgrade to a plan with same limits and a longer billing period
                current_period = customer.bill_period
                max_period_avail = plans['bill_period'].max()
                if current_period < max_period_avail and current_period < customer.max_bill_period:
                    first_limit = plans.columns.values[2]
                    same_limit_plans = plans[plans[first_limit]==customer.limits[first_limit]]
                    # the order is random - so customer could go to any higher period plan
                    higher_period_plans =  same_limit_plans[same_limit_plans['bill_period']>customer.bill_period].sample(frac=1)
                    new_plan_name = higher_period_plans.index.values[0]
                    customer.set_plan(plans,plan_name=new_plan_name)
                    changed_bill_period=True

        # if they didn't upgrade, check for downgrades
        if not changed_plan and uniform(0, 1) < downgrade_probability:
            # See if any downgrades are available
            if plans.shape[1] > 2:
                first_limit = plans.columns.values[2]
                same_period_plans = plans[plans['bill_period']==customer.bill_period]
                # sorting so the customer takes the next lower limit plan
                lower_limit_plans =  same_period_plans[same_period_plans[first_limit]<customer.limits[first_limit]].sort_values(by=first_limit)
                if len(lower_limit_plans)>0:
                    new_plan_name = lower_limit_plans.index.values[len(lower_limit_plans)-1]
                    customer.set_plan(plans,plan_name=new_plan_name)
                    changed_plan=True
            # If no suitable dowgrade, try to lower the billing period
            if not changed_plan and 'bill_period' in plans.columns.values:
                same_limit_plans = plans[plans[first_limit]==customer.limits[first_limit]]
                # the order is random - so customer could go to any lower period plan
                lower_period_plans =  same_limit_plans[same_limit_plans['bill_period']<customer.bill_period].sample(frac=1)
                if len(lower_period_plans)>0:
                    new_plan_name = lower_period_plans.index.values[len(lower_period_plans)-1]
                    customer.set_plan(plans,plan_name=new_plan_name)
                    changed_bill_period=True

        # add ons check - same as upgrade probability
        if uniform(0, 1) < upgrade_probability:
            shuffled_add_ons = add_ons.sample(frac=1).reset_index(drop=True)
            for add_on in shuffled_add_ons.iterrows():
                if len(customer.add_ons)>0 and add_on[1]['plan'] in customer.add_ons['plan'].values:
                    continue
                near_limit = False
                for limited in add_on[1].index.values[3:]:
                    customer_rate = customer.get_behavior_rate(limited)
                    if customer_rate > 0.5*customer.limits[limited] and add_on[1].loc[limited]>0:
                        near_limit = True
                        break
                if not near_limit:
                    continue
                if len(customer.add_ons)==0:
                    customer.add_ons=pd.DataFrame([add_on[1]])
                else:
                    add_on_df = pd.DataFrame([add_on[1]])
                    customer.add_ons = pd.concat([customer.add_ons,add_on_df])
                changed_add_ons=True
                break

        if not changed_add_ons and uniform(0, 1) < downgrade_probability:
            for add_on in customer.add_ons.iterrows():
                customer.add_ons = customer.add_ons.drop(customer.add_ons[customer.add_ons['plan']==add_on[1]['plan']].index)
                changed_add_ons=True
                break

        if changed_plan or changed_add_ons:
            customer.add_add_ons(plans)

        return changed_bill_period



import numpy as np
import os
import pandas as pd
import tempfile

from datetime import  datetime, timedelta,time
from dateutil import relativedelta
from filelock import FileLock
from math import ceil
from numpy import random
from random import randrange, randint


class Customer:
    date_multipliers = {}
    ID_FILE = os.path.join(tempfile.gettempdir(), f'churn_customer_id.txt')
    ID_LOCK_FILE = os.path.join(tempfile.gettempdir(), f'churn_customer_id_lock.txt')

    def __init__(self,behavior_rates,start_of_month,args,channel_name='NA'):
        '''
        Creates a customer for simulation, given an ndarray of behavior rates, which are converted to daily.
        Each customer also has a unique integer id which will become the account_id in the database, and holds its
        own subscriptions and events.

        For a detailed explanation see the ChurnSim report
        https://papers.ssrn.com/sol3/papers.cfm?abstract_id=4540160
        - Section 3.2.1, Behavior Model - basics of the behavior model
        - Section 3.3, Simulation Algorithm - use of the daily rates
        - Section 3.4.1, Day of Week Behavioral Fluctuation
        - Section 3.4.2, Product Channels
        - Section 3.4.3, Customer Satisfiability Coefficient
        - Section 3.4.4, Customer Age
        - Section 3.5.1, Multi-User Accounts
        - Section 3.5.2, Action Values
        - Section 3.5.6, Billing Period
        - Section 3.5.7. Discounts

        :param behavior_rates: ndarray of behavior rates, which are assumed to be PER MONTH
        :param start_of_month: The date of the month the customer starts
        :param args: OmegaConf with the program args
        :param channel_name: string name of the channel this customers behavior was drawn from
        '''
        with FileLock(Customer.ID_LOCK_FILE):
            next_id = 0
            if os.path.exists(Customer.ID_FILE):
                with open(Customer.ID_FILE, 'r') as id_file:
                    next_id = int(id_file.readline())
            self.id=next_id # set the id to the current class variable
            with open(Customer.ID_FILE, 'w') as id_file:
                id_file.write(f'{next_id+1}\n')
            if self.id % 100==0:
                print(f'Simulating customer {self.id}...')

        self.behavior_rates = behavior_rates
        self.behavior_rates['mean_value'] = None
        self.args=args


        for valued_behavior in Customer.get_valued_behaviors(self.behavior_rates['behavior'].values):
            underlying_behavior = Customer.get_behavior_under_value(valued_behavior, self.behavior_rates['behavior'].values)
            uidx= self.behavior_rates['behavior']==underlying_behavior
            vidx= self.behavior_rates['behavior']==valued_behavior
            self.behavior_rates.loc[uidx,'mean_value']= self.behavior_rates.loc[vidx,'monthly_rate'].values[0]
            self.behavior_rates = self.behavior_rates.drop(self.behavior_rates[vidx].index,axis=0)

        if 'users' in behavior_rates['behavior'].values:
            bidx= self.behavior_rates['behavior'] == 'users'
            self.users = max(int( round(self.behavior_rates.loc[bidx,'monthly_rate'])),1)
            self.behavior_rates = self.behavior_rates.drop(self.behavior_rates[bidx].index,axis=0)
        else:
            self.users = None


        self.behavior_rates['daily_rate'] = (1.0/30.0)*self.behavior_rates['monthly_rate']
        self.channel=channel_name

        age_range =self.args.max_age - self.args.min_age
        avg_age = age_range/2.0
        if start_of_month:
            self.age=random.uniform(self.args.min_age,self.args.max_age)
            self.date_of_birth = start_of_month + relativedelta.relativedelta(years=-int(self.age),
                                                                              months=-int( (self.age % 1)*12 ),
                                                                              days=-random.uniform(1,30))
        else:
            self.date_of_birth=None
            self.age = avg_age

        self.age_satisfaction_coef = self.args.age_satisfy * (avg_age - self.age)/age_range
        self.country="NA"
        self.mrr=None
        self.discount=0.0
        self.bill_period=1
        self.max_bill_period=None
        self.base_mrr=None
        self.plan=None
        self.add_ons = pd.DataFrame()
        self.limits= {}

        self.satisfaction_propensity = np.power(self.args.satisfy_base,
                                                random.uniform(-self.args.satisfy_scale, self.args.satisfy_scale) \
                                                + self.age_satisfaction_coef )

        self.subscriptions=[]
        self.events=[]
        self.current_utility = None
        self.utility_contribs = None

    def get_behavior_rate(self,behavior):
        if behavior == 'users':
            customer_rate = self.users
        elif behavior in self.behavior_rates['behavior'].values:
            customer_rate = self.behavior_rates[self.behavior_rates['behavior'] == behavior]['monthly_rate'].values[0]
        else:
            raise ValueError(f'get_behavior_rates UNKNOWN behavior {behavior}')
        return customer_rate

    @staticmethod
    def get_valued_behaviors(behavior_list):
        value_behaviors=[]
        for behave in behavior_list:
            if Customer.get_behavior_under_value(behave, behavior_list) is not None:
                value_behaviors.append(behave)
        return value_behaviors

    @staticmethod
    def get_underlying_behaviors(behavior_list):
        underlying_behaviors=[]
        for behave in behavior_list:
            if Customer.get_behavior_under_value(behave, behavior_list) is None:
                underlying_behaviors.append(behave)
        return underlying_behaviors

    @staticmethod
    def get_behavior_under_value(one_behavior, behavior_list):
        if not one_behavior.endswith('_value'):
            return None
        base_behave=str(one_behavior).replace('_value','')
        if base_behave in behavior_list:
            return base_behave
        else:
            return None

    def pick_initial_plan(self, plans, add_ons, bill_periods=None):
        """
        For a detailed explanation see the ChurnSim report
        https://papers.ssrn.com/sol3/papers.cfm?abstract_id=4540160
        - Section 3.5.3 Product Plans, MRR and Limits
        - Section 3.5.6 Billing Periods

        :param plans: Data frame of plans
        :param add_ons: Data frame of add-ons
        :param bill_periods: List of all the possible billing periods
        :return:
        """

        if bill_periods is not None:
            self.max_bill_period=np.random.choice(bill_periods)
            plans_to_use = plans[plans['bill_period'] <= self.max_bill_period]
        else:
            plans_to_use = plans

        if plans.shape[1] < 3:
            choice_index = np.random.choice(range(len(plans_to_use)))
        else:
            # Customer can pick between plans for which its expected number of users
            # is at least 1/3 the plan limit and at most 3x the plan limit
            eligible_plans = []
            for plan_index in range(0,len(plans_to_use)):
                is_eligible = False
                for limited in plans_to_use.columns.values[2:]:
                    plan_limit = plans_to_use.iloc[plan_index][limited]
                    customer_rate = self.get_behavior_rate(limited)
                    min_rate = 0.33*plan_limit if plan_index >0 else 0 # everyone is eligible_plans for first plan
                    max_rate = 3.0  * plan_limit
                    if min_rate <= customer_rate <= max_rate:
                        is_eligible = True
                        break
                if is_eligible:
                    eligible_plans.append(plan_index)
            assert len(eligible_plans)>0
            if len(eligible_plans)==1:
                choice_index = eligible_plans[0]
            else:
                choice_index = np.random.choice(eligible_plans)


        self.set_plan(plans_to_use,choice_index)
        if len(add_ons)>0:
            for add_on in add_ons.iterrows():
                if random.uniform(0,1) <= add_on[1]['prob']:
                    if len(self.add_ons)==0:
                        self.add_ons=pd.DataFrame([add_on[1]])
                    else:
                        new_add_on_df = pd.DataFrame([add_on[1]])
                        self.add_ons=pd.concat([self.add_ons,new_add_on_df])
        self.add_add_ons(plans_to_use)


    def set_plan(self,plans,plan_idx=None, plan_name=None):
        """
        For a detailed explanation see the ChurnSim report
        https://papers.ssrn.com/sol3/papers.cfm?abstract_id=4540160
        - Section 3.5.3, Product PLans, MRR and Limits
        - Section 3.5.5, Add-on Products
        - Section 3.5.6, Billing Periods

        :param plans: data frame of plans
        :param plan_idx: the plan index to set the plan to
        :param plan_name: a name of the plane to set to
        :return:
        """
        if plan_idx is not None:
            self.plan = plans.index.values[plan_idx]
        else:
            self.plan = plan_name
        # 50% chance of a discount as low as 50%
        if np.random.uniform(0, 1) <= self.args.discount_prob:
            # discounts come in multiples of the min, e.g. 1%, 2%, 5% - not any old value
            self.discount = self.args.min_discount * np.round(np.random.uniform(self.args.min_discount,self.args.max_discount)/0.05)
        else:
            self.discount=0.0
        self.mrr = np.round(plans.loc[self.plan,'mrr']* (1.0-self.discount))
        self.base_mrr = self.mrr
        if 'bill_period' in plans.columns.values:
            self.bill_period = plans.loc[self.plan,'bill_period']
        if plans.shape[1]>2:
            self.limits = {
                behave : plans.loc[self.plan, behave] for behave in plans.columns[2:]
            }

    def add_add_ons(self,plans):
        """
        For a detailed explanation see the ChurnSim report
        https://papers.ssrn.com/sol3/papers.cfm?abstract_id=4540160
        - Section 3.5.3, Product PLans, MRR and Limits
        - Section 3.5.5, Add-on Products

        :param plans: data frame of the plans
        :return:
        """
        # Reset to  base MRR
        self.set_plan(plans,plan_name=self.plan)
        for add_on in self.add_ons.iterrows():
            self.mrr+=add_on[1]['mrr']
            for limited in self.add_ons.columns[3:]:
                if limited in self.limits:
                    self.limits[limited]+= add_on[1][limited]
                else:
                    raise ValueError(f'Add on raises a limit {limited} that was not found')


    @staticmethod
    def get_min_max_dow_scale(scale_param):
        """
        For a detailed explanation see the ChurnSim report
        https://papers.ssrn.com/sol3/papers.cfm?abstract_id=4540160
        - Section 3.4.1, Day of Week Behavioral Fluctuation and equation 6

        :param scale_param: Scaling parameter for the day
        :return:
        """
        assert -1.0 <= scale_param < 1.0
        if scale_param > 0:
            min_scale = 1 - 0.1 * scale_param
            max_scale = 1 + scale_param
        else:
            min_scale = 1 + scale_param
            max_scale = 1  - 0.1 * scale_param
        return min_scale, max_scale


    @staticmethod
    def get_day_multiplier(the_date,args):
        """
        Increasing/Decreasing behavior based on the day of the week. Stores one multiplier to be
        used by all customers, the first time a day is encountered.

        For a detailed explanation see the ChurnSim report
        https://papers.ssrn.com/sol3/papers.cfm?abstract_id=4540160
        - Section 3.4.1, Day of Week Behavioral Fluctuation

        :param the_date:
        :param args:
        :return:
        """
        # Set a multiplier for this date
        if the_date not in Customer.date_multipliers:
            if the_date.weekday() >= 4:  # Friday - Sun
                min_scale, max_scale = Customer.get_min_max_dow_scale(args.weekend_scale)
            else:  # Monday-Thursday
                min_scale, max_scale = Customer.get_min_max_dow_scale(args.weekday_scale)
            multiplier = random.uniform(min_scale, max_scale)
            Customer.date_multipliers[the_date] = multiplier
        return Customer.date_multipliers[the_date]


    def generate_events(self,start_date,end_date):
        '''
        Generate a sequence of events at the customers daily rates.  Each count for an event on a day is droing from
        a poisson distribution with the customers average rate.  If the number is greater than zero, that number of events
        are created as tuples of time stamps and the event index (which is the database type id).  The time of the
        event is randomly set to anything on the 24 hour range.

        For a detailed explanation see the ChurnSim report
        https://papers.ssrn.com/sol3/papers.cfm?abstract_id=4540160
        - Section 3.2.1, Behavior Model
        - Section 3.4.1, Day of Week Behavioral Fluctuation
        - Section 3.5.1, Multi-User Accounts
        - Section 3.5.2, Action Values
        - Section 3.5.3, Product Plans, MRR and Limits (specifically the explanation of action limits)

        :param start_date: datetime.date for start of simulation
        :param end_date: datetime.date for end of simulation
        :return: The total count of each event, the list of all of the event tuples
        '''

        delta = end_date - start_date

        events=[]

        if self.users is None:
            counts=[0]*len(self.behavior_rates)
        else:
            counts=[0]*(len(self.behavior_rates)+1) # Plus one for number of users
        limit_counts = {b : 0 for b in self.limits.keys()}
        for i in range(delta.days):
            the_date = start_date + timedelta(days=i)
            dow_multiplier = Customer.get_day_multiplier(the_date,self.args)
            if self.users is None:
                todays_users = 1
            else:
                todays_users = int(round(dow_multiplier*random.poisson(self.users)))
                if 'users' in self.limits:
                    todays_users = min(todays_users,self.limits['users'])

            for row in  self.behavior_rates.iterrows():
                rate = row[1]['daily_rate']
                behavior_name = row[1]['behavior']
                new_count= int(round(dow_multiplier*random.poisson(rate) * todays_users))
                if behavior_name in self.limits:
                    # print(new_count, self.limits, limit_counts)
                    new_count = min(new_count, self.limits[behavior_name]-limit_counts[behavior_name])
                    limit_counts[behavior_name]=limit_counts[behavior_name]+new_count

                for n in range(0,new_count):
                    event_time=datetime.combine(the_date,time(randrange(24),randrange(60),randrange(60)))
                    user_id = 0
                    if self.users is not None:
                        user_id = randint(0, todays_users-1)
                    if row[1]['mean_value'] is not None:
                        event_value = round(np.exp(np.random.normal(np.log(row[1]['mean_value']) )),2)
                        counts[row[0]] += event_value
                    else:
                        event_value = None
                        counts[row[0]] += 1
                    new_event=(event_time, row[0], user_id, event_value)
                    events.append(new_event )

            if self.users is not None:
                counts[-1] += todays_users

        if self.users is not None:
            counts[-1] = int(ceil( counts[-1]/delta.days)) # user count is returned as average, not total

        self.events.extend(events)

        return counts


import os
import pandas as pd
import pickle
import shap
import matplotlib.pyplot as plt
from fightchurn.listings.chap8.listing_8_2_logistic_regression import prepare_data
from fightchurn.listings.chap8.listing_8_4_rescore_metrics import reload_churn_data

def shap_explain_xgb(data_set_path, plot_n=None, save_values=False):

    pickle_path = data_set_path.replace('.csv', '_xgb_model.pkl')
    assert os.path.isfile(pickle_path), 'You must run listing 9.6 to save an XGB regression model first'
    with open(pickle_path, 'rb') as fid:
        xgb_model = pickle.load(fid)

    # current_df = reload_churn_data(data_set_path,'current','8.3',is_customer_data=True)
    X, y = prepare_data(data_set_path, ext='', as_retention=False)
    explainer = shap.TreeExplainer(xgb_model,feature_perturbation = "interventional", model_output='probability',data=X)
    shap_values = explainer(X)
    shap.summary_plot(shap_values, X, show=False)
    save_file = data_set_path.replace('.csv', '_shap_summary_xgb.png')
    print(f'Saving SHAP Explanation to {save_file}')
    plt.tight_layout()
    plt.savefig(save_file, format='png')
    plt.close()

    if save_values:
        shap_data = pd.DataFrame(shap_values.values,columns=X.columns)
        save_path = data_set_path.replace('.csv', '_shapvals.csv')
        shap_data.to_csv(save_path,header=True)
        print('Saved shap values data to ' + save_path)
        shap_summary = pd.DataFrame(shap_data.describe()).transpose()
        save_path = data_set_path.replace('.csv', '_shapsummary.csv')
        shap_summary.to_csv(save_path,header=True)

    if plot_n is not None:
        for n in plot_n:
            shap.waterfall_plot(shap_values[n],show=False)
            save_file = data_set_path.replace('.csv', f'_shap_water_xgb_{n}.png')
            plt.tight_layout()
            plt.savefig(save_file, format='png')
            plt.close()



import pandas as pd
import numpy as np
import os

def metric_scores(data_set_path,skew_thresh=4.0):

    assert os.path.isfile(data_set_path),'"{}" is not a valid dataset path'.format(data_set_path)
    churn_data = pd.read_csv(data_set_path,index_col=[0,1])
    data_scores = churn_data.copy()
    data_scores.drop('is_churn',axis=1)

    stat_path = data_set_path.replace('.csv', '_summarystats.csv')
    assert os.path.isfile(stat_path),'You must running listing 5.2 first to generate stats'
    stats = pd.read_csv(stat_path,index_col=0)
    stats=stats.drop('is_churn')
    skewed_columns=(stats['skew']>skew_thresh) & (stats['min'] >= 0)
    skewed_columns=skewed_columns[skewed_columns]

    for col in skewed_columns.keys():
        data_scores[col]=np.log(1.0+data_scores[col])
        stats.at[col,'mean']=data_scores[col].mean()
        stats.at[col,'std']=data_scores[col].std()

    data_scores=(data_scores-stats['mean'])/stats['std']
    data_scores['is_churn']=churn_data['is_churn']

    score_save_path=data_set_path.replace('.csv','_scores.csv')
    print('Saving results to %s' % score_save_path)
    data_scores.to_csv(score_save_path,header=True)




from fightchurn.listings.chap8.listing_8_2_logistic_regression import prepare_data
from fightchurn.listings.chap9.listing_9_1_regression_auc  import reload_regression
import numpy

def calc_lift(y_true, y_pred):
    if numpy.unique(y_pred).size < 10:
        return 1.0
    sort_by_pred=[(p,t) for p,t in sorted(zip(y_pred, y_true))]
    overall_churn = sum(y_true)/len(y_true)
    i90=int(round(len(y_true)*0.9))
    top_decile_count=sum([p[1] for p in sort_by_pred[i90:]])
    top_decile_churn = top_decile_count/(len(y_true)-i90)
    lift = top_decile_churn/overall_churn
    return lift

def top_decile_lift(data_set_path):

    logreg_model = reload_regression(data_set_path)
    X,y = prepare_data(data_set_path,as_retention=False)
    predictions = logreg_model.predict_proba(X)
    lift = calc_lift(y,predictions[:,0])
    print('Regression Lift score={:.3f}'.format(lift))



import pandas as pd
import matplotlib.pyplot as plt
import os

def cohort_plot(data_set_path, metric_to_plot='',ncohort=10, exclude_zeros=False, log_scale=False):
    assert os.path.isfile(data_set_path),'"{}" is not a valid dataset path'.format(data_set_path)
    churn_data = pd.read_csv(data_set_path,index_col=[0,1])
    if exclude_zeros:
        churn_data = churn_data[churn_data[metric_to_plot]!=0]
    groups = pd.qcut(churn_data[metric_to_plot], ncohort, duplicates='drop')
    cohort_means = churn_data.groupby(groups)[metric_to_plot].mean()
    cohort_churns = churn_data.groupby(groups)['is_churn'].mean()
    plot_frame = pd.DataFrame({metric_to_plot: cohort_means.values, 'churn_rate': cohort_churns.values})
    plt.figure(figsize=(6, 4))
    plt.plot(metric_to_plot, 'churn_rate', data=plot_frame,marker='o', color='black', linewidth=2, label=metric_to_plot)
    plt.xlabel('Cohort Average of  "%s"' % metric_to_plot)
    plt.ylabel('Cohort Churn Rate')
    plt.grid()
    plt.gca().set_ylim(bottom=0)
    if log_scale:
        plt.gca().set_xscale('log')
    save_path = data_set_path.replace('.csv', '_' + metric_to_plot + '_churn_cohort.png')
    plt.savefig(save_path)
    print('Saving plot to %s' % save_path)



import pandas as pd
import matplotlib.pyplot as plt
import os

def metric_histogram(data_set_path, bins=None, metric_to_plot='',limits=[],log_x=False, log_y=False):
    assert os.path.isfile(data_set_path),'"{}" is not a valid dataset path'.format(data_set_path)
    churn_data = pd.read_csv(data_set_path,index_col=[0,1])
    plt.figure(figsize=(6, 4))
    plt.hist(churn_data[metric_to_plot],bins=bins)
    if limits:
        plt.xlim(limits)
    if log_x:
        plt.gca().set_xscale('log')
    if log_y:
        plt.gca().set_yscale('log')
    plt.title(metric_to_plot)
    plt.xlabel(f'mean={churn_data[metric_to_plot].mean():.0f}, '
               f'max={churn_data[metric_to_plot].max():.0f}, '
               f'skew={churn_data[metric_to_plot].skew():.1f}')
    plt.ylabel('# of Observations')
    plt.tight_layout()
    plt.grid()
    save_path = data_set_path.replace('.csv', '_' + metric_to_plot + '_histogram.png')
    plt.savefig(save_path)
    print('Saving plot to %s' % save_path)


import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="fightchurn",
    version="1.1.4",
    author="Carl Gold",
    author_email="carl@fightchurnwithdata.com",
    description="Code from the book Fighting Churn With Data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/carl24k/fight-churn",
    project_urls={
        "Bug Tracker": "https://github.com/carl24k/fight-churn/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_data={'fightchurn' : ['fightchurn/churnsim/*/*.csv',
                                  'fightchurn/churnsim/*/*.sql',
                                  'fightchurn/churnsim/*/*.yaml',
                                  'fightchurn/listings/conf/*.json',
                                  'fightchurn/listings/*/*.sql']},
    include_package_data=True,
    packages=['fightchurn',
              'fightchurn.churnsim',
              'fightchurn.churnsim.conf',
              'fightchurn.churnsim.schema',
              'fightchurn.listings',
              'fightchurn.listings.conf',
              'fightchurn.listings.chap3',
              'fightchurn.listings.chap5',
              'fightchurn.listings.chap6',
              'fightchurn.listings.chap7',
              'fightchurn.listings.chap8',
              'fightchurn.listings.chap9',
              'fightchurn.listings.chap10'],
    scripts=['fightchurn/run_churn_listing.py',
             'fightchurn/churnsim/churndb.py',
             'fightchurn/churnsim/churnsim.py'],
    python_requires=">=3.9,<3.10",
    install_requires=["antlr4-python3-runtime==4.9.3.0",
                        "bleach==6.0.0",
                        "build==0.10.0",
                        "certifi==2023.7.22",
                        "chardet==5.2.0",
                        "cloudpickle==2.2.1",
                        "colorama==0.4.6",
                        "cycler==0.11.0",
                        "docutils==0.20.1",
                        "greenlet==2.0.2",
                        "hydra-core==1.3.2",
                        "importlib-metadata==6.8.0",
                        "joblib==1.3.2",
                        "keyring==24.2.0",
                        "kiwisolver==1.4.4",
                        "llvmlite==0.40.1",
                        "matplotlib==3.7.2",
                        "numba==0.57.1",
                        "numpy==1.24.4",
                        "packaging==23.1",
                        "pandas==1.5.3",
                        "patsy==0.5.3",
                        "Pillow==10.0.1",
                        "filelock==3.12.2",
                        "omegaconf==2.3.0",
                        "pkginfo==1.9.6",
                        "postgres==4.0",
                        "psycopg2-binary==2.9.7",
                        "psycopg2-pool==1.1",
                        "Pygments==2.16.1",
                        "pyparsing==2.4.7",
                        "python-dateutil==2.8.2",
                        "pytz==2023.3",
                        "pytest",
                        "readme-renderer==41.0",
                        "requests==2.31.0",
                        "requests-toolbelt==1.0.0",
                        "rfc3986==1.5.0",
                        "scikit-learn==1.3.0",
                        "scipy==1.11.2",
                        "shap==0.42.1",
                        "six==1.16.0",
                        "slicer==0.0.7",
                        "SQLAlchemy==1.4.9",
                        "statsmodels==0.14.0",
                        "toml==0.10.2",
                        "tqdm==4.66.1",
                        "twine==4.0.2",
                        "urllib3==2.0.7",
                        "webencodings==0.5.1",
                        "xgboost==1.7.6",
                        "zipp==3.16.2"]
)



import pandas as pd
import numpy as np
from fightchurn.listings.chap7.listing_7_5_fat_tail_scores import transform_fattail_columns, transform_skew_columns
from fightchurn.listings.chap8.listing_8_4_rescore_metrics import reload_churn_data

def clip_hi_cols(data, hi_vals):
    for col in hi_vals.index.values:
        data.loc[data[col] > hi_vals[col],col] = hi_vals[col]

def clip_lo_cols(data, lo_vals):
    for col in lo_vals.index.values:
        data.loc[data[col] < lo_vals[col],col] = lo_vals[col]

def rescore_metrics(data_set_path):

    current_data = reload_churn_data(data_set_path,'current','8.3',is_customer_data=True)
    load_mat_df = reload_churn_data(data_set_path,'load_mat','6.4',is_customer_data=False)
    score_df = reload_churn_data(data_set_path,'score_params','7.5',is_customer_data=False)
    stats = reload_churn_data(data_set_path,'summarystats','5.2',is_customer_data=False)
    stats.drop('is_churn',inplace=True)
    assert set(score_df.index.values)==set(current_data.columns.values),"Data to re-score does not match transform params"
    assert set(load_mat_df.index.values)==set(current_data.columns.values),"Data to re-score does not match load matrix"
    assert set(stats.index.values)==set(current_data.columns.values),"Data to re-score does not match summary stats"

    clip_hi_cols(current_data, stats['99pct'])
    clip_lo_cols(current_data, stats['1pct'])

    transform_skew_columns(current_data, score_df[score_df['skew_score']].index.values)
    transform_fattail_columns(current_data, score_df[score_df['skew_score']].index.values)

    current_data=current_data[score_df.index.values]
    scaled_data=(current_data-score_df['mean'])/score_df['std']

    scaled_data = scaled_data[load_mat_df.index.values]
    grouped_ndarray = np.matmul(scaled_data.to_numpy(), load_mat_df.to_numpy())

    current_data_grouped = pd.DataFrame(grouped_ndarray,columns=load_mat_df.columns.values, index=current_data.index)

    score_save_path=data_set_path.replace('.csv','_current_groupscore.csv')
    current_data_grouped.to_csv(score_save_path,header=True)
    print('Saving results to %s' % score_save_path)


